import os
import sys
from timeit import default_timer as timer
from gams import GamsWorkspace
from gams.database import GamsDatabase
from gams import GamsVariable

# Read execution parameters
args = sys.argv
assert len(args) > 2, "GAMS path and input path must be specified."
ws = GamsWorkspace(system_directory=args[1])
input_path: str = os.path.abspath(args[2])
assert input_path[-4:] == ".gdx", f"{input_path} is not a .gdx file."
dir_path: str = os.path.split(input_path)[0]

db: GamsDatabase = ws.add_database_from_gdx(input_path)
output: GamsDatabase = ws.add_database()

# Find the dummy variables by searching for symbols starting with 'exo_' or 'endo_'
exo_dummies = [symbol for symbol in db if symbol.name[:4] == "exo_"]
endo_dummies = [symbol for symbol in db if symbol.name[:5] == "endo_"]

endo_exo_strings = []

for dummy in exo_dummies:
	var_name = dummy.name[4:]
	if len(db.get_symbol(var_name)) > 1:
		endo_exo_strings += [f"{var_name}.fx{rec.keys} = {var_name}.l{rec.keys};" for rec in dummy]
	else:
		endo_exo_strings += f"{var_name}.fx = {var_name}.l;"

for dummy in endo_dummies:
	var_name = dummy.name[5:]
	endo_exo_strings += [f"{var_name}.lo{rec.keys} = -inf; {var_name}.up{rec.keys} = inf;" for rec in dummy]

gms_path = os.path.join(dir_path, 'include_endo_exo.gms')
with open(gms_path, "w") as file:
	file.write("\n".join(endo_exo_strings))


# def test_endo_exo():
# 	args = ["", "C:\\GAMS\\win64\\28.1", "C:\\Users\\B031441\\Desktop\\HEAD\\Model\\Saved\\sim_input.gdx"]



import xlwings as xw
import logging

from dreamtools import unstack_multiseries

from dreamtools.gams_pandas import Gdx
import os


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

ACTIVE_GDX = None


@xw.func
@xw.arg('path', doc='Path of GDX file.')
def set_gdx_path(path):
  """Set file path of active GDX file."""
  global ACTIVE_GDX
  ACTIVE_GDX = Gdx(path)
  logger.info(f"ACTIVE_GDX set to {ACTIVE_GDX.abs_path}")
  return f"Active gdx set to {ACTIVE_GDX.abs_path}"


@xw.func
def get_gdx_path():
  """Get file path of currently active GDX file."""
  return getattr(ACTIVE_GDX, "abs_path", None)


@xw.arg('request', doc='Name of variable.')
@xw.arg('year', doc='Optional. Year to get data.')
@xw.arg('gdx_path', doc='Optional. File path of gdx file to get data from.')
@xw.arg('array_results', doc='Optional, default=False. If True, the function can return an array with values from multiple years or array series elements.')
@xw.func
@xw.ret(header=True, index=True, expand='table')
def gdx(request, year=None, gdx_path=None, array_results=False):
  """
  Look up the request in a gdx file.
  A request uses Gekko syntax, e.g. "qBNP", "qY[tje]", or "qY[#s]".
  If no year is specified, an array is returned with values for all available years.
  If no gdx_path is specified, use the global path set using set_gdx_path
  """
  if gdx_path is not None:
    gdx = Gdx(gdx_path)
  elif ACTIVE_GDX is not None:
    gdx = ACTIVE_GDX
  else:
    return "#ERROR! Use set_gdx_path to choose a gdx file."

  var_name, records = parse_request(gdx, request)
  if var_name not in gdx:
    return "#VALUE!"
  result = gdx[var_name]
  if records:
    result = result.loc[records]
  result = unstack_multiseries(result, keep_axis_index=-1)
  if year:
    result = result.loc[year].values
  if len(result) > 1 and not array_results:
    return "#VALUE!"
  else:
    return result


def parse_request(gdx, name):
  if name.count("[") == 1 and name.endswith("]"):
    var_name = name.split("[")[0]
    domains = name.split("[")[1][:-1].split(",")
    return var_name, tuple(parse_domain(gdx, domain) for domain in domains)
  else:
    return name, []


def parse_domain(gdx, domain):
  if domain.startswith("#"):
    return ["".join(element) for element in gdx[domain[1:]].index]  # "".join(element) 'de-tuples' the index element
  else:
    try:
      return [int(domain)]
    except ValueError:
      return [domain]



# @xw.func
# @xw.ret(header=False, index=False, expand='table')
# def test(x):
#   # return str(gams.GamsWorkspace(system_directory="C:\\GAMS\\win64\\28.2", debug=3))
#   # return str(gams.GamsWorkspace())
#   return(str(x))



"""
Transform Excel work sheet into GAMS code to perform endo exo operations and change variable levels.
"""
import os
import sys
import pandas as pd
import xlwings as xw


def main():
    #  Read execution paremeters
    workbook_path = os.path.abspath(sys.argv[1])
    assert os.path.splitext(workbook_path)[1].lower() in [".xls", ".xlsx", ".xlsm"], \
        f"{workbook_path} is not an Excel workbook"

    output_path = os.path.abspath(sys.argv[2])

    # Read Excel file
    print(f"Read endo_exo sheet from {workbook_path}")
    app = xw.App(visible=False)
    workbook = xw.Book(workbook_path)
    endo_exo = workbook.sheets["endo_exo"].range("B4:CK5000").options(pd.DataFrame, index=False).value
    app.quit()

    # Cast years in column header to ints
    header = list(endo_exo)
    year_columns_index = 4
    header[year_columns_index:] = [int(i) for i in header[year_columns_index:]]
    endo_exo.columns = header

    endo_exo_strings = endo_exo.apply(row_to_gekko, axis=1, result_type="reduce", year_columns_index=year_columns_index)

    with open(output_path, "w") as file:
        print(f"Write to {output_path}")
        file.write("\n".join(endo_exo_strings))


def row_to_gekko(row, year_columns_index):
    """Return Gekko code from row"""
    endo_exo_strings = []
    for interval in split_intervals(row[year_columns_index:]):
        if row['#']:  # Transfer comments
            endo_exo_strings += [f"// {i}" for i in row.dropna()]
        elif interval:
            interval_string = f"{interval[0]} {interval[-1]}"
            if row.endo:
                endo_exo_strings += [f"ENDO <{interval_string}> {row.endo};"]
            if row.exo:
                endo_exo_strings += [f"EXO <{interval_string}> {row.exo};"]
                values = "(" + ", ".join(row.loc[interval].astype(str)) + ")"
                if row.printcode == "dummy":
                    pass
                elif row.printcode:
                    endo_exo_strings += [f"<{interval_string} {row.printcode}> {row.exo} = {values};"]
                else:
                    endo_exo_strings += [f"<{interval_string}> {row.exo} = {values};"]

    return "\n".join(endo_exo_strings)


def split_name(input):
    """
    Split Gekko syntax name
    Return symbol name and a list of elements
    Example:
        >>> split_name("qI[#i,CON]")
        "qI", ["#i", "'CON'"]
    """
    if "[" in input:
        name, elements = input.split("[", 1)
        elements = elements[:-1].split(",")
        elements = [e[1:] if (e[0] == "#") else f"'{e}'" for e in elements]
    else:
        name = input
        elements = []
    return name, elements


def split_intervals(series):
    interval = []
    for year in series.dropna().index:
        if (year - 1) not in interval:
            if interval:
                yield interval
                interval = []
        interval += [year]
    yield interval


def error(msg):
    exit(f"ERROR! {msg}")


if __name__ == "__main__":
    main()



import plotly.graph_objects as go

dream_colors_rgb = [
  "rgb(245,82,82)",   # Red (DREAM)
  "rgb(20,175,166)",  # Pretrolium (MAKRO)
  "rgb(255,155,75)",  # Orange (SMILE)
  "rgb(92,210,114)",  # Green (Grøn REFORM)
  "rgb(66,180,224)",  # Blue (REFORM)
  "rgb(188,173,221)", # Plum
  "rgb(0,95,151)",    # Dark blue
  "rgb(137,48,112)",  # Maroon
  "rgb(70,70,76)",    # Dark gray
  "rgb(230,230,232)", # Light gray
]

dream_colors_hex = [
  "F55252",
  "14AFA6",
  "FF9B4B",
  "5CD272",
  "42B4E0",
  "BCADDD",
  "005F97",
  "893070",
  "46464C",
  "E6E6E8",
]

dream = {
  "layout": {
    "colorway": dream_colors_rgb,
    "legend": {
      "yanchor": "bottom",
      "xanchor": "center",
      "orientation": "h",
      "y": -0.5,
      "x": 0.5,
    },
    "font": {"color": "rgb(0,0,0)", "family": "Hind", "size": 14},

    "xaxis": {
      "ticks": "outside",
      "ticklen": 10,
      "showline": True,
      "linecolor": "black",
      "automargin": False,
      "title_standoff": 0,
      "title_font_size": 14,
    },

    "yaxis": {
      "ticks": "outside",
      "ticklen": 10,
      "showline": True,
      "linecolor": "black",
      # "tickangle": -90,
      "automargin": False,
      "tickmode": "auto",
    },

    "plot_bgcolor": "rgb(230,230,232)",
    "xaxis_gridcolor": "white",
    "yaxis_gridcolor": "white",

    # "paper_bgcolor": "rgb(230,230,232)",
  },

  "data": {
    "scatter": [{
      "line": {"width": 4},
    }],
  }
}

dream["data"]["scattergl"] = dream["data"]["scatter"]

dream_template = go.layout.Template(**dream)


"""
"""
import os

import numpy as np
import pandas as pd
import gams
import logging
import builtins
import time
from copy import deepcopy
from .utility import index_from_symbol, symbol_is_scalar, is_iterable, map_lowest_level, index_names_from_symbol, all_na
import gams2numpy

logger = logging.getLogger(__name__)

class GamsPandasDatabase:
  """
  GamsPandasDatabase converts sets, parameters, and variables between a GAMS database and Pandas series.
  When as symbol is first retrieved it is converted to a Pandas series and stored in self.series
  Changes to retrieved series are written to the GAMS database on export.
  """

  def __init__(self, database=None, workspace=None, auto_sort_index=True):
    if database is None:
      if workspace is None:
        workspace = gams.GamsWorkspace()
      database = workspace.add_database()
    self.database = database
    self.gams2numpy = gams2numpy.Gams2Numpy(database.workspace.system_directory)
    self.auto_sort_index = auto_sort_index
    self.series = {}

  def __getattr__(self, item):
    return self[item]

  def copy(self):
    obj = type(self).__new__(self.__class__)
    obj.__dict__.update(self.__dict__)
    obj.database = self.database.workspace.add_database(source_database=self.database)
    obj.series = deepcopy(self.series)
    return obj

  def merge(self, other, symbol_names=None, inplace=False):
    """
    Merge two GamsPandasDatabases.
    symbol_names: list of symbol names to get from other. If None, all symbols in other are merged.
    """
    if inplace:
      db = self
    else:
      db = self.copy()

    if symbol_names is None:
      symbol_names = other.symbols

    for name in other.sets:
      if name in symbol_names:
        series = other[name].texts
        db.create_set(series.name, series.index, other[name].explanatory_text, series)

    for name in other.variables:
      if name in symbol_names:
        db.add_variable_from_series(other[name], other[name].explanatory_text)

    for name in other.parameters:
      if name in symbol_names:
        db.add_parameter_from_series(other[name], other[name].explanatory_text)

    return db

  @property
  def symbols(self):
    """Dictionary of all symbols in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database}

  @property
  def sets(self):
    """Dictionary of all sets in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsSet)}

  @property
  def variables(self):
    """Dictionary of all variables in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsVariable)}

  @property
  def parameters(self):
    """Dictionary of all parameters in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsParameter)}

  @property
  def equations(self):
    """Dictionary of all equations in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsEquation)}

  def add_to_builtins(self, *args):
    """Retrieve any number symbol names from the database and add their Pandas representations to the global namespace."""
    for identifier in args:
      setattr(builtins, identifier, self[identifier])

  def get(self, *args):
    """Retrieve any nymber of symbol names and return a list of their Pandas representations."""
    return [self[i] for i in args]

  def add_parameter_from_dataframe(self, identifier, df, explanatory_text="", add_missing_domains=False,
                                   value_column_index=-1):
    """Add parameter symbol to database based on a Pandas DataFrame."""
    domains = list(df.columns[:value_column_index:])
    for d in domains:
      if d not in self:
        if add_missing_domains:
          self.create_set(name=d, index=df[d].unique(), explanatory_text="")
        else:
          raise KeyError(
            f"'{d}' is not a set in the database. Enable add_missing_domains or add the set to the database manually.")
    self.database.add_parameter_dc(identifier, domains, explanatory_text)
    if domains:
      series = df.set_index(domains).iloc[:, 0]
    else:
      series = df[df.columns[0]]
    self.series[identifier] = series

  def add_parameter_from_series(self, series, explanatory_text="", add_missing_domains=False):
    """Add parameter symbol to database based on a Pandas series."""
    if len(series) == 1 and series.index.name in [None, "*"]:
      df = pd.DataFrame(series)
    else:
      df = series.reset_index()
    self.add_parameter_from_dataframe(series.name, df, explanatory_text, add_missing_domains)

  def add_variable_from_dataframe(self, identifier, df, explanatory_text="", add_missing_domains=False,
                                  value_column_index=-1):
    """Add variable symbol to database based on a Pandas DataFrame."""
    domains = list(df.columns[:value_column_index:])
    for d in domains:
      if d not in self:
        if add_missing_domains:
          self.create_set(name=d, index=df[d].unique(), explanatory_text="")
        else:
          raise KeyError(
            f"'{d}' is not a set in the database. Enable add_missing_domains or add the set to the database manually.")
    self.database.add_variable_dc(identifier, gams.VarType.Free, domains, explanatory_text)
    if domains:
      series = df.set_index(domains).iloc[:, 0]
    else:
      series = df[df.columns[0]]
    self.series[identifier] = series

  def add_variable_from_series(self, series, explanatory_text="", add_missing_domains=False):
    """Add variable symbol to database based on a Pandas series."""
    if len(series) == 1 and series.index.name in [None, "*"]:
      df = pd.DataFrame(series)
    else:
      df = series.reset_index()
    self.add_variable_from_dataframe(series.name, df, explanatory_text, add_missing_domains)

  def get_index(self, x):
    if x is None or isinstance(x, pd.Index):
      return x
    elif isinstance(x, str):
      return self[x]
    elif len(x) and isinstance(x[0], (pd.Index, tuple, list)):
      multi_index = pd.MultiIndex.from_product(x)
      multi_index.names = [getattr(i, "name", None) for i in x]
      return multi_index
    else:
      return pd.Index(x)

  def create_set(self, name, index, explanatory_text="", texts=None, domains=None):
    """
    Add a new GAMS Set to the database and return an Pandas representation of the Set.
    :param str name: Name of the set
    :param iterable index: Iterable of record keys to be added to the set
    :param str explanatory_text: Explanatory text added to the GAMS set
    :param iterable texts: Iterable of record labels - should match the size of the index parameter.
    :param iterable domains: Names of domains that the set should be defined over
    :return: Pandas Index
    """
    if len(index) and isinstance(index[0], pd.Index):
      multi_index = pd.MultiIndex.from_product(index)
      multi_index.names = [getattr(i, "name", None) for i in index]
      index = multi_index
    elif isinstance(index, pd.Index):
      index = index.copy()
    else:
      index = pd.Index(index)

    index.explanatory_text = explanatory_text

    if texts is None:
      texts = map_lowest_level(str, index)
    index.texts = pd.Series(texts, index=index)
    index.texts.name = index.name

    if domains is None:
      domains = ["*" if i in (None, name) else i for i in self.get_domains_from_index(index, name)]
    index.domains = domains
    index.names = domains
    index.name = name

    self.database.add_set_dc(index.name, domains, explanatory_text)
    self.series[index.name] = index
    return self[name]

  def __create_variable_or_parameter(self, symbol_type, name, index, explanatory_text, data, dtype, copy, add_missing_domains):
    if index is not None:
      series = pd.Series(data, self.get_index(index), dtype, name, copy)
      series.explanatory_text = explanatory_text
      getattr(self, f"add_{symbol_type}_from_series")(series, explanatory_text, add_missing_domains)
    elif isinstance(data, pd.DataFrame):
      getattr(self, f"add_{symbol_type}_from_dataframe")(name, data, explanatory_text, add_missing_domains)
    elif isinstance(data, pd.Series):
      getattr(self, f"add_{symbol_type}_from_series")(data, explanatory_text, add_missing_domains)
    else:
      if is_iterable(data) and len(data) and is_iterable(data[0]):
        self.__add_variable_or_parameter_to_database(symbol_type, name, len(data[0]), explanatory_text)
      elif is_iterable(data):
        getattr(self, f"add_{symbol_type}_from_dataframe")(name, pd.DataFrame(data), explanatory_text, add_missing_domains)
      else:
        self.__add_variable_or_parameter_to_database(symbol_type, name, 0, explanatory_text)
        self[name] = data
    return self[name]

  def __add_variable_or_parameter_to_database(self, symbol_type, name, dim, explanatory_text):
    assert symbol_type in ["parameter", "variable"]
    if symbol_type == "parameter":
      self.database.add_parameter(name, dim, explanatory_text)
    elif symbol_type == "variable":
      self.database.add_variable(name, dim, gams.VarType.Free, explanatory_text)

  def create_variable(self, name, index=None, explanatory_text="", data=None, dtype=None, copy=False, add_missing_domains=False):
    return self.__create_variable_or_parameter("variable", name, index, explanatory_text, data, dtype, copy, add_missing_domains)

  def create_parameter(self, name, index=None, explanatory_text="", data=None, dtype=None, copy=False, add_missing_domains=False):
    return self.__create_variable_or_parameter("parameter", name, index, explanatory_text, data, dtype, copy, add_missing_domains)

  @staticmethod
  def get_domains_from_index(index, name):
    if hasattr(index, "domains"):
      domains = index.domains
    elif hasattr(index, "name"):
      domains = index.names
    else:
      domains = [index.name]
    return ["*" if i in (None, name) else i for i in domains]

  @staticmethod
  @np.vectorize
  def detuple(t):
    """Returns the iterable unchanged, except if it is a singleton, then the element is returned"""
    if isinstance(t, str):
      return t
    try:
      if len(t) == 1:
        return t[0]
    except TypeError:
      pass
    return t

  def series_from_symbol(self, symbol, attributes, attribute):
    index_names = index_names_from_symbol(symbol)
    df = pd.DataFrame(
      self.gams2numpy.gmdReadSymbolStr(self.database, symbol.name),
      columns=[*index_names, *attributes],
    )
    for i in index_names:
      df[i] = df[i].astype(int, errors="ignore")
    series = df.set_index(index_names)[attribute].astype(float)
    series.name = symbol.name
    return series

  def series_from_variable(self, symbol, attr="level"):
    return self.series_from_symbol(symbol, attributes=["level", "marginal", "lower", "upper", "scale"], attribute="level")

  def series_from_parameter(self, symbol):
    return self.series_from_symbol(symbol, attributes=["level"], attribute="level")

  def __getitem__(self, item):
    if item not in self.series:
      symbol = self.symbols[item]

      if isinstance(symbol, gams.GamsSet):
        self.series[item] = index_from_symbol(symbol)

      elif isinstance(symbol, gams.GamsVariable):
        if symbol_is_scalar(symbol):
          self.series[item] = symbol.find_record().level if len(symbol) else None
        else:
          self.series[item] = self.series_from_variable(symbol)
          # self.series[item] = series_from_variable(symbol)
          if self.auto_sort_index:
            self.series[item] = self.series[item].sort_index()

      elif isinstance(symbol, gams.GamsParameter):
        if symbol_is_scalar(symbol):
          self.series[item] = symbol.find_record().value if len(symbol) else None
        else:
          self.series[item] = self.series_from_parameter(symbol)
          # self.series[item] = series_from_parameter(symbol)
          if self.auto_sort_index:
            self.series[item] = self.series[item].sort_index()

      elif isinstance(symbol, gams.GamsEquation):
        return symbol

    return self.series[item]

  def __setitem__(self, name, value):
    if name in self.symbols:
      if not is_iterable(value) and is_iterable(self[name]):  # If assigning a scalar to all records in a series
        value = pd.Series(value, index=self[name].index)
      self.set_symbol_records(self.symbols[name], value)
      self.series[name] = value
    else:
      if not value.name:
        value.name = name
      self.series[name] = value

  def items(self):
    return self.symbols.items()

  def keys(self):
    return self.symbols.keys()

  def values(self):
    return self.symbols.values()

  def save_series_to_database(self, series_names=None):
    """Save Pandas series to GAMS database"""
    if series_names is None:
      series_names = self.series.keys()
    for symbol_name in series_names:
      self.set_symbol_records(self.symbols[symbol_name], self.series[symbol_name])

  def export(self, path):
    """Save changes to database and export database to GDX file."""
    self.save_series_to_database()
    for i in range(1, 10):
      try:
        self.database.export(os.path.abspath(path))
        break
      except gams.workspace.GamsException:
        time.sleep(i)
    else:
      self.database.export(os.path.abspath(path))

  def set_parameter_records(self, symbol, value):
    symbol.clear()
    if all_na(value): pass
    elif symbol_is_scalar(symbol):
      symbol.add_record().value = value
    else:
      df = value.reset_index()
      df[value.index.names] = df[value.index.names].astype(str)
      self.gams2numpy.gmdFillSymbolStr(self.database, symbol, df.to_numpy())

  @staticmethod
  def set_variable_records(symbol, value):
    symbol.clear()
    if all_na(value): pass
    elif symbol_is_scalar(symbol):
      symbol.add_record().level = value
    elif list(value.keys()) == [0]:  # If singleton series
      symbol.add_record().level = value[0]
    else:
      for k, v in value.items():
        symbol.add_record(map_lowest_level(str, k)).level = v

  @staticmethod
  def set_set_records(symbol, value):
    if isinstance(value, pd.Index):
      texts = getattr(value, "texts", None)
      value = texts if texts is not None else pd.Series(map(str, value), index=value)

    symbol.clear()
    if all_na(value): pass
    elif symbol_is_scalar(symbol):
      symbol.add_record().text = str(value)
    elif list(value.keys()) == [0]:  # If singleton series
      symbol.add_record().text = value[0]
    else:
      for k, v in value.items():
        symbol.add_record(map_lowest_level(str, k)).text = v

  def set_symbol_records(self, symbol, value):
    """Convert Pandas series to records in a GAMS Symbol"""
    if isinstance(symbol, gams.GamsSet):
      self.set_set_records(symbol, value)
    elif isinstance(symbol, gams.GamsVariable):
      self.set_variable_records(symbol, value)
    elif isinstance(symbol, gams.GamsParameter):
      self.set_parameter_records(symbol, value)
    else:
      TypeError(f"{type(symbol)} is not (yet) supported by gams_pandas")

  def __iter__(self):
    return iter(self.symbols)

  def __len__(self):
    return len(self.symbols)

  def get_text(self, name):
    """Get explanatory text of GAMS symbol."""
    return self.symbols[name].get_text()



import setuptools

with open("README.md", "r") as file:
  long_description = file.read()

setuptools.setup(
  name="dream-tools",
  version="0.5.2",
  author="Martin Kirk Bonde",
  author_email="martin@bonde.dk",
  description="A collection of tools used by the Danish institute for economic modelling and forecasting, DREAM (http://dreammodel.dk).",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/MartinBonde/dream-tools",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
  ],
  python_requires='>=3.6',
  install_requires=["pandas", "easygui"],
)


import dreamtools as dt
import numpy as np
import pandas as pd
from IPython.display import display


def time(start, end=None):
  """Set global time settings."""
  if end is None:
    end = start
  dt.START_YEAR = start
  dt.END_YEAR = end

def years():
  return list(range(dt.START_YEAR, dt.END_YEAR+1))

def aggregate_index(index, default_set_aggregations):
  if index.nlevels > 1:
    return tuple(default_set_aggregations.get(k, list(index.levels[i])) for i, k in enumerate(index.names))
  else:
    return default_set_aggregations.get(index.name, list(index))

def aggregate_series(series, default_set_aggregations=None, reference_database=None):
  if default_set_aggregations is None:
    default_set_aggregations = dt.DEFAULT_SET_AGGREGATIONS

  if reference_database is None:
    reference_database = get_reference_database()

  if series.name in reference_database:
    if len(reference_database[series.name].index) == len(series.index):
      return series.loc[aggregate_index(series.index, default_set_aggregations)]
  elif series.index.name in reference_database:
    if len(reference_database[series.index.name]) == len(series.index):
      return series.loc[aggregate_index(series.index, default_set_aggregations)]

  return series

def prt(iter_series,
        operator=None,
        start_year=None,
        end_year=None,
        reference_database=None,
        default_set_aggregations=None,
        function=None,
        dec=6,
        max_rows=100,
        max_columns=20,):
  df = to_dataframe(iter_series, operator, start_year, end_year, reference_database, default_set_aggregations, function)
  df.style.set_properties(**{"text-align": "right", "precision": dec})
  with pd.option_context('display.max_rows', max_rows, 'display.max_columns', max_columns):  # more options can be specified also
    display(df)

def add_xline(fig, x):
  "Add a vertical line to a plotly figure at x"
  return fig.update_layout(shapes=[dict(
    type='line',
    yref='paper', y0=0, y1=1,
    xref='x', x0=x, x1=x,
    line=dict(
      dash="dash",
    ),
    opacity=0.3,
  )])

def plot(iter_series,
         operator=None,
         start_year=None,
         end_year=None,
         reference_database=None,
         default_set_aggregations=None,
         function=None,
         names=None,
         xline=None,
         layout={},
         **kwargs):
  df = to_dataframe(iter_series, operator, start_year, end_year, reference_database, default_set_aggregations, function)
  if names:
    df.columns = names
  fig = df.plot(**kwargs)
  layout = {
    "yaxis_title_text": dt.YAXIS_TITLE_FROM_OPERATOR.get(operator, ""),
    "xaxis_title_text": dt.TIME_AXIS_TITLE,
    "legend_title_text": "",
    **layout
  }
  fig.update_layout(layout)
  if xline is not None:
    fig = add_xline(fig, xline)
  return fig

def to_dataframe(iter_series,
                 operator=None,
                 start_year=None,
                 end_year=None,
                 reference_database=None,
                 default_set_aggregations=None,
                 function=None):
  if isinstance(iter_series, pd.Series):
    iter_series = [iter_series]

  if function is None:
    function = lambda x: x

  iter_series = [function(aggregate_series(s, default_set_aggregations, reference_database))
                 for s in iter_series]
  if operator:
    if reference_database is None:
      reference_database = get_reference_database()
    refs = [function((s - s + reference_database[s.name])[s.index])
            if not dimension_changed(s, reference_database)
            else s * np.NaN
            for i, s in enumerate(iter_series)]
    iter_series = compare(iter_series, refs, operator)

  if start_year is None:
    start_year = dt.START_YEAR
  if end_year is None:
    end_year = dt.END_YEAR

  return merge_multiseries(*iter_series).loc[start_year:end_year]

def dimension_changed(series, reference_database):
  if series.name in reference_database:
    is_changed = reference_database[series.name].index.nlevels != series.index.nlevels
    if is_changed:
      Warning(
        f"The dimension of '{series.name}' is different in the reference database. If indexing a single element write [['element']] rather than ['element'] to prevent the series dimension being reduced.")
    return is_changed
  else:
    Warning(f"'{series.name}' was not found in the reference database.")
    return True


def get_reference_database():
  if dt.REFERENCE_DATABASE is None:
    import easygui
    dt.REFERENCE_DATABASE = dt.Gdx(easygui.fileopenbox("Select reference gdx file", filetypes=["*.gdx"]))
  return dt.REFERENCE_DATABASE

def compare(iter_series, refs, operator):
  """
  Applies an operator to each pair in zip(iter_series, refs)
  """
  if operator in ["q"]:
    return [s / b - 1 for s, b in zip(iter_series, refs)]
  elif operator in ["pq"]:
    return [(s / b - 1)*100 for s, b in zip(iter_series, refs)]
  elif operator in ["m"]:
    return [s - b for s, b in zip(iter_series, refs)]
  elif operator in ["s"]:
    for b in refs:
      b.name = "baseline." + b.name
    return [i for pair in zip(iter_series, refs) for i in pair]
  elif operator in ["p"]:
    return [s / lag(s) * 100 - 100 for s in iter_series]
  elif operator in ["d"]:
    return [s - lag(s) for s in iter_series]
  elif operator in ["i"]:
    return [index(s) for s in iter_series]
  elif operator in ["log"]:
    return [np.log(s) for s in iter_series]
  elif operator in ["rlog"]:
    return [np.log(s) for s in refs]
  elif operator in ["dlog"]:
    return [np.log(s) - np.log(lag(s)) for s in iter_series]
  else:
    raise ValueError(f"{operator} is not a valid operator.")

def index(series, level=-1, element=None):
  """Return series divided by the value with the index element. E.g. set a series to 1 in a base year."""
  if element is None:
    element = dt.START_YEAR
  if len(series.index.names) > 1:
    return series / series.xs(element, level=level)
  else:
    return series / series[element]

def lag(series, periods=1, lag_axis_index=-1):
  if len(series.index.names) > 1:
    return series.groupby(level=series.index.names[:lag_axis_index:]).shift(periods)
  else:
    return series.shift(periods)

def unstack_multiseries(series, keep_axis_index=-1):
  """
  Return a DataFrame from a Series.
  All levels of the Series are unstacked and concatenated as column names, except <keep_axis_index>.
  """
  if isinstance(series.index, pd.MultiIndex):
    series = series.copy()
    if series.name is None:
      series.name = ""
    series.index = pd.MultiIndex.from_tuples(flatten_keys(series.name, keys, keep_axis_index) for keys in series.index)
    df = series.unstack(0)[series.index.get_level_values(0).unique()]
  else:
    df = pd.DataFrame(series)
  return df

def flatten_keys(name, keys, keep_axis_index):
  keys_str = ','.join(map(str, keys[:keep_axis_index:]))
  flat_name = f"{name}[{keys_str}]"
  return flat_name, keys[keep_axis_index]

def merge_multiseries(*series, keep_axis_index=-1):
  """
  Return a DataFrame from any number of Series, with all levels except <keep_axis_index> concatenated as column names.
  """
  output = pd.DataFrame()
  for s in series:
    df = unstack_multiseries(s, keep_axis_index)
    for c in df.columns:
      new_name = c if (c != "0") else ""
      iter = 0
      while new_name in output:
        iter += 1
        new_name = f"{c}{iter}"
      output[new_name] = df[c]
  return output



from math import ceil

from plotly import graph_objects as go, offline as pyo
from plotly.subplots import make_subplots

import pandas as pd

import dreamtools as dt

def small_figure(fig):
    "Adjust figure with settings suitable for side by side use in the Office suite"
    legend_height = 200 / dt.PLOT_SCALE + len(fig.data) * 64 / dt.PLOT_SCALE
    width, height = dt.SMALL_PLOT_WIDTH / dt.PLOT_SCALE, dt.PLOT_HEIGHT / dt.PLOT_SCALE
    fig = go.Figure(fig).update_layout(
        width = width,
        height = height + legend_height,
        legend_y = - 200 / dt.PLOT_HEIGHT,
        legend_yanchor = "top",
        legend_orientation = "v",
        margin = {"t": 20, "b": legend_height, "l": 60, "r": 10},

        # Title is used as y-axis label
        yaxis_title_text = "",
        title_text = fig.layout.yaxis.title.text,
        title_xanchor = "left", title_x = 0,
        title_yanchor = "top", title_y = 1,
        title_pad_l = 7, title_pad_t = 7,
        title_font_size = 14,
    )
    if  max(len(i.name) for i in fig.data) > 30:  # Long legend entries cause the plot area size to change when centered
        fig.update_layout(legend_x = 0, legend_xanchor = "left")
    return fig

def large_figure(fig):
    "Adjust figure with settings suitable for use in the Office suite"
    trace_count = len(fig.data)
    col_count = 2
    row_count = trace_count // col_count
    for i, trace in enumerate(fig.data):
        trace.legendgroup = i // (trace_count / col_count)
    legend_height = 200 / dt.PLOT_SCALE + row_count * 64 / dt.PLOT_SCALE
    width, height = dt.LARGE_PLOT_WIDTH / dt.PLOT_SCALE, dt.PLOT_HEIGHT / dt.PLOT_SCALE
    fig = go.Figure(fig).update_layout(
        width = width,
        height = height + legend_height,
        legend_y = - 200 / dt.PLOT_HEIGHT,
        legend_yanchor = "top",
        margin = {"t": 20, "b": legend_height, "l": 60, "r": 10},

        # Title is used as y-axis label
        yaxis_title_text="",
        title_text=fig.layout.yaxis.title.text,
        title_xanchor="left", title_x=0,
        title_yanchor="top", title_y=1,
        title_pad_l=7, title_pad_t=7,
        title_font_size=14,
    )
    return fig

def figures_to_html(figs, filename="figures.html"):
  """Write an iter of plotly figures to an html file."""
  file = open(filename, 'w')
  file.write("<html><head></head><body>" + "\n")

  for i, fig in enumerate(figs):
    inner_html = pyo.plot(
      fig, include_plotlyjs=(i==0), output_type='div'
    )
    file.write(inner_html)

  file.write("</body></html>" + "\n")

def subplot(figures, cols=2, **kwargs):
  """Create subplot from iter of figures."""
  rows = ceil(len(figures) / cols)
  fig = make_subplots(rows=rows,
                      cols=cols,
                      subplot_titles=[f["layout"]["title"]["text"] for f in figures],
                      **kwargs
                      )
  for i, f in enumerate(figures):
    for trace in f["data"]:
      fig.add_trace(trace, row=ceil((i+1)/cols), col=1 + i % cols)
  return fig

def html_table(df, precision=2, header=True, header_background_color="#14AFA6", head_text_color="#ffffff"):
  """Return string with a formated HTML version of a Pandas Dataframe"""
  df = df.copy()
  df.index.name, df.columns.name = None, None

  styles = [
    # table properties
    dict(selector=" ",
         props=[
           ("margin", "0"),
           ("font-family", '"Hind", "Helvetica", "Arial", sans-serif'),
           ("border-collapse", "collapse"),
           # ("border", "2px solid #0F837D"),
           ("text-align", "right"),
         ]),

    # header color
    dict(selector="thead",
         props=[
           ("background-color", header_background_color),
           ("color", head_text_color),
         ]),

    # background shading
    dict(selector="tbody",
         props=[("background-color", "#E6E6E8")]),
    # dict(selector="tbody tr:nth-child(even)",
    #         props=[("background-color", "#fff")]),
    # dict(selector="tbody tr:nth-child(odd)",
    #         props=[("background-color", "#eee")]),

    # body cell properties
    dict(selector="td",
         props=[
           ("font-size", "120%"),
           ("padding", ".5em"),
           ("border", "1px solid #ffffff"),
         ]),

    # header cell properties
    dict(selector="th",
         props=[
           ("font-size", "120%"),
           ("text-align", "right"),
           ("padding", ".5em"),
           ("border", "1px solid #ffffff"),
         ]),

    # caption placement
    dict(selector="caption",
         props=[("caption-side", "bottom")]),
  ]
  if not header:
    styles.append({"selector": "thead", "props": [("font-size", "0pt")]})

  with pd.option_context("precision", precision):
    return "<meta charset='UTF-8'>" + df.style.set_table_styles(styles).render()